/*
 * Copyright (c) 2006, WSO2 Inc. (http://www.wso2.org) All Rights Reserved.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package org.wso2.registry.jdbc.dao;

import org.wso2.registry.Resource;
import org.wso2.registry.LogEntry;
import org.wso2.registry.RegistryConstants;
import org.wso2.registry.RegistryException;
import org.wso2.registry.jdbc.DatabaseConstants;
import org.apache.commons.logging.Log;
import org.apache.commons.logging.LogFactory;

import java.sql.*;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;

public class VersionedResourceDAO {

    private static final Log log = LogFactory.getLog(VersionedResourceDAO.class);
    /**
     * Returns the given version of the artifact. If artifact is a collection all child paths will
     * be added as versioned paths in the form /a/b?v=12
     *
     * @param path : path of the artifact
     * @return Resource for the given version number
     * @throws SQLException : If something went wrong
     * @param versionNumber  : Version number
     * @param conn : Connection to DB
     */
    public Resource get(String path, long versionNumber, Connection conn) throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(path + "?v=" + versionNumber);
            resource.setParentPath(getParentPath(resource.getId(), versionNumber, conn));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (resource == null) {
            return null;
        }

        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource chidren

        if (resource.isDirectory()) {

            String[] children = getChildPaths(resource.getId(), versionNumber, conn);
            resource.setContent(children);
        }

        // get resource properties

        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, resource.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        resource.setProperties(props);

        return resource;
    }

    public Resource getResourceByID(long artifactID, long versionNumber, Connection conn) throws SQLException {

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.AID=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setLong(1, artifactID);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (resource == null) {
            return null;
        }

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(artifactID, conn);
        }
        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, versionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }

        // get resource chidren

        if (resource.isDirectory()) {

            String[] children = getChildPaths(resource.getId(), versionNumber, conn);
            resource.setContent(children);
        }

        // get resource properties

        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, resource.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        resource.setProperties(props);

        return resource;
    }

    /**
     * Returns the latest (current) version of an artifact. If artifact is a collection, all child
     * paths will be added as latest paths in the form /a/b
     *
     * @param path : path of the artifacy
     * @param conn :Connection to DB
     * @return : Result artifcat
     * @throws SQLException : if somthing went wrong
     */
    public Resource getLatestVersion(String path, Connection conn) throws SQLException {
        /*
        * when someone give path like "root/wso2/wsas/r1" , then need to figure out the latest
        * version of the root and from the need to figure out the version of wso2 and so so
        * so when list outing the nodes need to give them correctly.
        * */

        Resource resource = null;

        ResultSet result;

        // get the resource data from the resource table

        String sql = "SELECT * FROM ARTIFACTS A WHERE A.PATH=?";
        PreparedStatement statement = conn.prepareStatement(sql);
        statement.setString(1, path);

        result = statement.executeQuery();

        if (result.next()) {
            resource = new Resource();

            resource.setId(result.getLong(DatabaseConstants.AID_FIELD));
            resource.setPath(result.getString(DatabaseConstants.PATH_FIELD));
            resource.setParentPath(getParentPath(resource.getPath()));
            resource.setDirectory(result.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
            resource.setMediaType(result.getString(DatabaseConstants.MEDIA_TYPE_FIELD));
            resource.setState(result.getInt(DatabaseConstants.STATE_FIELD));
            resource.setAuthorUserName(result.getString(DatabaseConstants.AUTHOR_FIELD));
            resource.setCreatedTime(result.getTimestamp(DatabaseConstants.CREATED_TIME_FIELD));
            resource.setDescription(result.getString(DatabaseConstants.DESCRIPTION_FIELD));
        }

        if (resource == null) {
            return null;
        }

//        long latestVersionNumber = getLatestVerisonNumber(resource.getId(), conn);
        long latestVersionNumber = getCurrentVerisonNumber(resource.getPath(), conn);


        String sqlVersionedFields = "SELECT * FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement versionQuery = conn.prepareStatement(sqlVersionedFields);
        versionQuery.setLong(1, resource.getId());
        versionQuery.setLong(2, latestVersionNumber);

        ResultSet versionResults = versionQuery.executeQuery();

        if (versionResults.next()) {
            resource.setLastUpdaterUserName(versionResults.getString(DatabaseConstants.VERSION_AUTHOR_FIELD));
            resource.setLastModified(versionResults.getTimestamp(DatabaseConstants.VERSION_UPDATED_TIME));
            resource.setContent(versionResults.getBytes(DatabaseConstants.VERSION_CONTENT_FIELD));
        }
        // get resource chidren
        if (resource.isDirectory()) {
            String[] children = getLatestChildPaths(resource.getId(),
                    getCurrentVerisonNumber(path, conn), conn);
            resource.setContent(children);
        }

        // get resource properties
        String propSQL = "SELECT P.PKEY, P.PVALUE FROM PROPERTIES P WHERE P.AID=?";
        PreparedStatement s3 = conn.prepareStatement(propSQL);
        s3.setLong(1, resource.getId());
        ResultSet propResults = s3.executeQuery();

        Properties props = new Properties();
        while (propResults.next()) {
            String key = propResults.getString(DatabaseConstants.PROPERTY_KEY_FIELD);
            String value = propResults.getString(DatabaseConstants.PROPERTY_VALUE_FIELD);
            props.put(key, value);
        }
        resource.setProperties(props);
        return resource;
    }

    /**
     * To get the currenet version for a give resource , current version may not be the latest
     * version in all the caess. If someone retore a resource then the child nodes will remain same
     * But the parent version will be changed. In that case the parent [foo/bar] verion will be the latest
     * version but the child node verion [foo/bar/v1] will be an old verion . So when some one ask for
     * the resource value foo/bar/v1 then we need to calculate them correctly and give that.
     * @param path : Resource ID
     * @param conn : Connection to DB
     * @return : verion number of the resource user requested
     */
    private long getCurrentVerisonNumber(String  path, Connection conn) {
        if (path != null) {
            try {
                int index = path.lastIndexOf(RegistryConstants.PATH_SEPARATOR);
                if (index >0) {
                    String parentPath = path.substring(0,index);
                    long parentId = getResourceID(parentPath,conn);
                    long latestVersion = getLatestVerisonNumber(parentId,conn);
                    long childID = getResourceID(path,conn);
                    String SQL = "SELECT DVN FROM DEPENDENCY WHERE AID=? AND VN=? AND DAID=?";
                    PreparedStatement s = conn.prepareStatement(SQL);
                    s.setLong(1, parentId);
                    s.setLong(2, latestVersion);
                    s.setLong(3, childID);
                    ResultSet result = s.executeQuery();
                    if (result.next()) {
                        return result.getLong(1);
                    }
                } else {
                   return getLatestVerisonNumber(getResourceID(path,conn),conn);
                }
            } catch (SQLException e) {
                log.info("Something went wrong while calculating the current " +
                        "version for the path : " + path+ " and the error us " + e);
                return -1;
            }
        }
        return -1;
    }

    public String[] getChildPaths(long artifcatID, long versionNumber,
                                  Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, artifcatID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            long childVersion = result.getLong("DVN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH") + "?v=" + childVersion;
                dependencyPathList.add(path);
            }
        }
        return (String[]) (dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    public String getParentPath(long resourceID, long versionNumber,
                                Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(resourceID, connection);
        }
        if (versionNumber == -1) {
            return null;
        }
        String selectSql = "SELECT AID , VN FROM DEPENDENCY WHERE DAID=? AND DVN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, resourceID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        String parentPath = "";

        //to see whether the patent node is there in the table , if not need to add that
        if (result.next()) {
            long aid = result.getLong("AID");
            long parentVersionNumber = result.getLong("VN");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, aid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                parentPath = path + "?v=" + parentVersionNumber;
            }
        }
        return parentPath;
    }

    public String[] getLatestChildPaths(long resourceID, long versionNumber,
                                        Connection connection) throws SQLException {

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, resourceID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList dependencyPathList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {
            long daid = result.getLong("DAID");
            String SQL = "SELECT PATH FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);
            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                String path = pathResult.getString("PATH");
                dependencyPathList.add(path);
            }
        }
        return (String[]) (dependencyPathList.toArray(new String[dependencyPathList.size()]));
    }

    public List getChildren(long resourceID, long versionNumber,
                            Connection connection) throws SQLException {

        if (versionNumber == -1) {
            versionNumber = getLatestVerisonNumber(resourceID, connection);
        }

        String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
        PreparedStatement s = connection.prepareStatement(selectSql);
        s.setLong(1, resourceID);
        s.setLong(2, versionNumber);
        ResultSet result = s.executeQuery();

        ArrayList childList = new ArrayList();

        //to see whether the patent node is there in the table , if not need to add that
        while (result.next()) {

            long daid = result.getLong("DAID");
            String SQL = "SELECT AID, PATH, DIRECTORY FROM ARTIFACTS WHERE AID=?";
            PreparedStatement ps = connection.prepareStatement(SQL);
            ps.setLong(1, daid);

            ResultSet pathResult = ps.executeQuery();
            if (pathResult.next()) {
                Resource child = new Resource();
                child.setId(pathResult.getLong(DatabaseConstants.AID_FIELD));
                child.setPath(pathResult.getString(DatabaseConstants.PATH_FIELD));
                child.setDirectory(pathResult.getInt(DatabaseConstants.DIRECTORY_FIELD) == 1);
                childList.add(child);
            }
        }
        return childList;
    }

    /**
     * Adds the given resource to the database.
     *
     * @param resource New resource
     * @param conn     : Connection to Db
     * @throws java.sql.SQLException : if something went wrong
     */
    public void add(Resource resource, Connection conn) throws SQLException {

        // first add the resource data into the artifacts table

        String sql = "INSERT INTO ARTIFACTS (PATH, MEDIA_TYPE, DIRECTORY, STATE, CREATED_TIME, AUTHOR, LAST_UPDATER, LAST_UPDATED_TIME,  DESCRIPTION, CONTENT)" +
                "VALUES (?,?,?,?,?,?,?,?,?,?)";

        long now = System.currentTimeMillis();

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, resource.getPath());
        s.setString(2, resource.getMediaType());
        s.setInt(3, resource.isDirectory() ? 1 : 0);
        s.setInt(4, RegistryConstants.ACTIVE_STATE);
        s.setTimestamp(5, new Timestamp(now));
        s.setString(6, resource.getAuthorUserName());
        s.setString(7, resource.getLastUpdaterUserName());
        s.setTimestamp(8, new Timestamp(now));
        s.setString(9, resource.getDescription());

        if (resource.getContent() != null && resource.getContent() instanceof byte[]) {
            s.setBytes(10, (byte[]) resource.getContent());
        } else {
            s.setBytes(10, null);
        }

        s.executeUpdate();


        long newArtifactID = this.getResourceID(resource.getPath(), conn);

        // now add resource properties to the properties table
        addProperties(newArtifactID, resource.getProperties(), conn);
    }

    public void update(String path, Resource resource, Connection conn) throws SQLException {

        // first update the artifacts table

        String sql = "UPDATE ARTIFACTS SET MEDIA_TYPE=?, DIRECTORY=?, STATE=?, AUTHOR=?, LAST_UPDATER=?, LAST_UPDATED_TIME=?, DESCRIPTION=?, CONTENT=? WHERE PATH=?";

        long now = System.currentTimeMillis();

        PreparedStatement s = conn.prepareStatement(sql);

        s.setString(1, resource.getMediaType());
        s.setInt(2, resource.isDirectory() ? 1 : 0);
        s.setInt(3, resource.getState());
        s.setString(4, resource.getAuthorUserName());
        s.setString(5, resource.getLastUpdaterUserName());
        s.setTimestamp(6, new Timestamp(now));
        s.setString(7, resource.getDescription());

        if (resource.getContent() != null && resource.getContent() instanceof byte[]) {
            s.setBytes(8, (byte[]) resource.getContent());
        } else {
            s.setBytes(8, null);
        }

        s.setString(9, path);

        s.executeUpdate();

        // remove existing properties for the updated artfact
        removeProperties(resource.getId(), conn);

        // now add the updated properties to the properties table
        addProperties(resource.getId(), resource.getProperties(), conn);
    }

    public void markDeleted(String path, Connection conn) throws SQLException {

        String sql = "UPDATE ARTIFACTS SET STATE=? WHERE PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setInt(1, RegistryConstants.DELETED_STATE);
        s.setString(2, path);
        s.executeUpdate();
    }

    public void addProperties(long resourceID, Properties props, Connection conn)
            throws SQLException {

        String propSQL = "INSERT INTO PROPERTIES (AID, PKEY, PVALUE) VALUES (?,?,?)";
        PreparedStatement s2 = conn.prepareStatement(propSQL);

        if (props != null) {

            Iterator i = props.keySet().iterator();
            while (i.hasNext()) {
                String key = (String) i.next();
                String value = (String) props.get(key);

                s2.setLong(1, resourceID);
                s2.setString(2, key);
                s2.setString(3, value);

                s2.executeUpdate();
                s2.clearParameters();
            }
        }
    }

    public void removeProperties(long resourceID, Connection conn) throws SQLException {

        String removePropsSQL = "DELETE FROM PROPERTIES WHERE AID=?";
        PreparedStatement s2 = conn.prepareStatement(removePropsSQL);
        s2.setLong(1, resourceID);
        s2.executeUpdate();
    }

    public boolean resourceExist(String path, Connection conn) throws SQLException {

        String sql = "SELECT COUNT(A.AID) FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        long count = 0;
        if (result.next()) {
            count = result.getLong(1);
        }

        return count == 1;
    }

    public long getResourceID(String path, Connection conn) throws SQLException {

        String sql = "SELECT A.AID FROM ARTIFACTS A WHERE A.PATH=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setString(1, path);

        ResultSet result = s.executeQuery();
        if (result.next()) {
            return result.getLong(DatabaseConstants.AID_FIELD);
        }

        return -1;
    }

    public String getResourcePath(long id, Connection conn) throws SQLException {

        String sql = "SELECT A.PATH FROM ARTIFACTS A WHERE A.AID=?";

        PreparedStatement s = conn.prepareStatement(sql);
        s.setLong(1, id);

        ResultSet result = s.executeQuery();
        if (result.next()) {
            return result.getString(DatabaseConstants.PATH_FIELD);
        }
        return null;
    }

    public void addResourceVersion(Resource resource, Connection connection) throws SQLException {

        if (resource.getLastModified() == null) {
            resource.setLastModified(new java.util.Date());
        }

        long versionNumber = insertToVersionTable(resource, connection);

        //updating max version table
        updateDependency(resource, versionNumber, false, connection);
    }

    private long insertToVersionTable(Resource resource, Connection connection)
            throws SQLException {

        String sql = "INSERT INTO VERSIONS (AID, VN ,CONTENT , AUTHOR, UPDATED_TIME) VALUES (?,?,?,?,?)";
        PreparedStatement statement = connection.prepareStatement(sql);
        statement.setLong(1, resource.getId());

        // Need to getthe old version and increment that
        // Selecet VID from latest table ,
        // if it is there then find from that  , if it is not there then we need ti start from 0
        long versionNumber = getLatestVerisonNumber(resource.getId(), connection);
        if (versionNumber == -1) {
            versionNumber = 1;
        } else {
            //incrementing the version number
            versionNumber = versionNumber + 1;
        }
        statement.setLong(2, versionNumber);

        Object content = resource.getContent();
        if (content != null && content instanceof byte[]) {
            statement.setBytes(3, (byte[]) content);
        } else {
            statement.setBytes(3, null);
        }

        statement.setString(4, resource.getLastUpdaterUserName());
        statement.setTimestamp(5, new Timestamp(System.currentTimeMillis()));
        statement.executeUpdate();

        return versionNumber;
    }

    private void updateDependency(Resource resource,
                                  long versionNumber,
                                  boolean delete, Connection connection) throws SQLException {

        // updating dependency table , so that we can rollback to a particular version
        // updating for the current resource
        long parentNextVersionNumber =
                updateDependencyTable(resource, versionNumber, delete, connection);

        // a/b/c/r1.wsdl
        // Next a/b/c
        // Next a/b
        // Next a
        if (parentNextVersionNumber != -1) {

            // updating its parent
            String path = resource.getPath();
            while (path != null && path.indexOf("/") > -1) {

                int lastIndex = path.lastIndexOf("/");
                String parentPath = path.substring(0, lastIndex);
                Resource parentArtifact = getLatestVersion(parentPath, connection);

                if (parentArtifact != null) {
                    path = parentArtifact.getPath();
                    parentNextVersionNumber = updateDependencyTable(
                            parentArtifact, parentNextVersionNumber, false, connection);

                    if (parentNextVersionNumber == -1) {
                        break;
                    }
                } else {
                    break;
                }
            }
        }
    }

    /**
     * @param resource      : Artifcat we need to update itslef and its immediate parent
     * @param versionNumber : Version number of the artifcat
     * @param delete        : delete or adding
     * @param connection    : DB connetion
     * @return : Will retuen the resource version of the parent resource
     * @throws SQLException : if something went wrong
     */
    private long updateDependencyTable(Resource resource,
                                       long versionNumber,
                                       boolean delete, Connection connection) throws SQLException {
        /**
         *        D0
         *    D1      D2
         *  D3  R1    R2
         *
         *  First fine D1 V0
         */
        // Checking whether the path is null or top most directory
        // Checking whether resource is root node or not

        String path = resource.getPath();

        // Checking whether the path is null or top most directory
        // if (path != null && path.indexOf("/") > -1) {
        if (path != null && !path.equals(RegistryConstants.ROOT_PATH)) {

            int lastIndex = path.lastIndexOf("/");
            String parentPath = RegistryConstants.ROOT_PATH;
            if (lastIndex != 0) {
                parentPath = path.substring(0, lastIndex);
            }

            if ("".equals(parentPath)) {
                //TODO : Need to think the logic here

            } else {

                long parentId = getResourceID(parentPath, connection);

                if (resource.getLastModified() == null) {
                    resource.setLastModified(new java.util.Date());
                }
                // Need to get the parent's parent and update the and update the

                long artifactversionNumber = getCurrentVerisonNumber(parentPath, connection);
//                long artifactversionNumber = getLatestVerisonNumber(parentId, connection);

                long nextVersionNumber = createNextVersion(parentId,
                        resource.getLastUpdaterUserName(),
                        resource.getLastModified().getTime(), connection);

                String selectSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
                PreparedStatement s = connection.prepareStatement(selectSql);
                s.setLong(1, parentId);
                s.setLong(2, artifactversionNumber);
                ResultSet result = s.executeQuery();

                ArrayList dependeyMap = new ArrayList();

                // to see whether the parent node is there in the table , if not need to add that
                while (result.next()) {

                    long daid = result.getLong("DAID");

                    if (daid != resource.getId()) {
                        long dvn = result.getLong("DVN");
                        PreparedStatement ps = getStatementForDependencyTable(connection,
                                parentId, nextVersionNumber, daid, dvn);
                        dependeyMap.add(ps);
                    }
                }

                if (!delete) {
                    PreparedStatement ds = getStatementForDependencyTable(connection,
                            parentId, nextVersionNumber, resource.getId(), versionNumber);
                    dependeyMap.add(ds);
                }

                for (int i = 0; i < dependeyMap.size(); i++) {
                    PreparedStatement preparedStatement = (PreparedStatement) dependeyMap.get(i);
                    preparedStatement.executeUpdate();
                }

                return nextVersionNumber;
            }

        } else {
            // TODO , I think we do not need to add the entry here , since when adding a child node,
            // if the parent is not there then we will add that
        }

        return -1;
    }

    public long getLatestVerisonNumber(long resourceId, Connection connection) throws SQLException {

        String sql = "SELECT MAX(VN) FROM VERSIONS WHERE AID=?";
        PreparedStatement s = connection.prepareStatement(sql);
        s.setLong(1, resourceId);
        ResultSet result = s.executeQuery();

        if (result.next()) {
            return result.getLong(1);
        }

        return -1;
    }

    public String[] getVersionNumbers(long resourceId, Connection connection) throws SQLException {

        String SQL = "SELECT VN FROM VERSIONS WHERE AID=?";
        PreparedStatement s = connection.prepareStatement(SQL);
        s.setLong(1, resourceId);
        ResultSet result = s.executeQuery();
        ArrayList versionNumbers = new ArrayList();
        while (result.next()) {
            versionNumbers.add("" + result.getLong("VN"));
        }
        return (String[]) versionNumbers.toArray(new String[versionNumbers.size()]);
    }

    public void restore(long resourceId, long toVersion, Connection connection) throws SQLException {

        Resource oldArtifact = getResourceByID(resourceId, toVersion, connection);

        String plainPath = oldArtifact.getPath();
        if (oldArtifact.getPath().indexOf("?v=") != -1) {
            String[] parts = oldArtifact.getPath().split("\\?v=");
            plainPath = parts[0];
        }

        update(plainPath, oldArtifact, connection);
        addResourceVersion(oldArtifact, connection);

        if (oldArtifact.isDirectory()) {

            String selectChildSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
            PreparedStatement selectps = connection.prepareStatement(selectChildSql);
            selectps.setLong(1, resourceId);
            selectps.setLong(2, toVersion);
            ResultSet childResult = selectps.executeQuery();
            ArrayList dependencyList = new ArrayList();

            while (childResult.next()) {

                long daid = childResult.getLong("DAID");
                long dvn = childResult.getLong("DVN");
                long versionNumber = getLatestVerisonNumber(resourceId, connection);
                PreparedStatement ps = getStatementForDependencyTable(connection,
                        resourceId, versionNumber, daid, dvn);
                dependencyList.add(ps);
            }

            for (int i = 0; i < dependencyList.size(); i++) {
                PreparedStatement preparedStatement = (PreparedStatement) dependencyList.get(i);
                preparedStatement.executeUpdate();
            }
        }
    }

    /**
     * To restoreVersion or rollback artifcat to a given version number
     *
     * @param resourceId : Resource ID
     * @param connection : Connection to DB
     * @param toVersion  : to which version we need to restoreVersion
     * @throws SQLException : If something went wrong
     */
    public void restore2(long resourceId, long toVersion, Connection connection) throws SQLException {

        String SQL = "SELECT CONTENT, AUTHOR, UPDATED_TIME FROM VERSIONS WHERE AID=? AND VN=?";
        PreparedStatement statement = connection.prepareStatement(SQL);
        statement.setLong(1, resourceId);
        statement.setLong(2, toVersion);

        ResultSet result = statement.executeQuery();
        long versionNumber = 0;
        if (result.next()) {
            Resource artifactToRestore = new Resource();
            artifactToRestore.setContent(result.getObject(1));
            artifactToRestore.setAuthorUserName(result.getString(2));
            artifactToRestore.setLastModified(result.getDate(3));
            versionNumber = insertToVersionTable(artifactToRestore, connection);
        }

        Resource resource = getResourceByID(resourceId, toVersion, connection);

        // updating max version table
        updateDependency(resource, versionNumber, false, connection);
        if (resource.isDirectory()) {
            String selectChildSql = "SELECT DAID, DVN FROM DEPENDENCY WHERE AID=? AND VN=?";
            PreparedStatement selectps = connection.prepareStatement(selectChildSql);
            selectps.setLong(1, resourceId);
            selectps.setLong(2, toVersion);
            ResultSet childResult = selectps.executeQuery();
            ArrayList dependencyList = new ArrayList();
            while (childResult.next()) {
                long daid = childResult.getLong("DAID");
                long dvn = childResult.getLong("DVN");
                PreparedStatement ps = getStatementForDependencyTable(connection,
                        resourceId, versionNumber, daid, dvn);
                dependencyList.add(ps);
            }

            for (int i = 0; i < dependencyList.size(); i++) {
                PreparedStatement preparedStatement = (PreparedStatement) dependencyList.get(i);
                preparedStatement.executeUpdate();
            }
        }

    }

    /**
     * To delete a resource , what actualy happen here is that update the version table with the new data
     * and resource will not actualy removed from the table
     *
     * @param resource   : Resource to be deleted
     * @param connection : Connection to the DB
     * @throws java.sql.SQLException : if something went wrong
     */
    public void deleteResource(Resource resource,
                               Connection connection) throws SQLException {
        long versionNumber = getLatestVerisonNumber(resource.getId(), connection);
        updateDependency(resource, versionNumber, true, connection);
    }

    private long createNextVersion(long resourceID,
                                   String lastUpdaterUserName,
                                   long lastModifiedTime,
                                   Connection connection) throws SQLException {

        long maxVersion = getLatestVerisonNumber(resourceID, connection);
        if (maxVersion == -1) {
            maxVersion = 0;
        }

        // here maxVersion is always >0
        maxVersion = maxVersion + 1;

        Resource resource = new Resource();
        resource.setId(resourceID);
        resource.setLastUpdaterUserName(lastUpdaterUserName);
        resource.setLastModified(new Date(lastModifiedTime));
        insertToVersionTable(resource, connection);

        return maxVersion;
    }

    private PreparedStatement getStatementForDependencyTable(Connection connection,
                                                             long parentId,
                                                             long nextVersionNumber,
                                                             long daid,
                                                             long dvn) throws SQLException {

        String SQL = "INSERT INTO DEPENDENCY (AID, VN , DAID , DVN) VALUES (?,? ,?,?)";
        PreparedStatement ps = connection.prepareStatement(SQL);
        ps.setLong(1, parentId);
        ps.setLong(2, nextVersionNumber);
        ps.setLong(3, daid);
        ps.setLong(4, dvn);
//        System.out.println("Adding dependency: " + parentId + " " + nextVersionNumber + " " + daid + " " + dvn);
        return ps;
    }

    /**
     * This method is use to see whether a give resource is actve or not , one can delte a resource
     * and try to access that giving the path of the resource in that case we should not return the
     * resource if it is deleted.
      * @param resourcePath : Path of the resource
     * @param connection : Connection to DB
     * @return : return true if the resource is active , else false
     */
    public boolean isResourceActive(String resourcePath ,Connection connection ){
        String[] parts = resourcePath.split(RegistryConstants.PATH_SEPARATOR);
        String currentPath = "";
        int current =0;
        if (parts.length > 0) {
            while (current < parts.length) {
                currentPath = currentPath + RegistryConstants.PATH_SEPARATOR + parts[current];
                if (!getResourceStatus(resourcePath,connection)){
                    return false;
                }
                current ++;
            }
        }
        return true;
    }

    /**
     * Will check whether the status of the given resource is active in the DB .,
     * this method differ from isResourceActive b'coz this is actulay call the DB
     * Not the other
     * @param resourcePath  : path to resource
     * @param connection : connection to DB
     * @return : If something went wrong
     */
    private boolean getResourceStatus(String resourcePath ,Connection connection ){
        try {
            String SQL = "SELECT STATE FROM ARTIFACTS WHERE PATH=?";
            PreparedStatement stateStm = connection.prepareStatement(SQL);
            stateStm.setString(1, resourcePath);
            ResultSet results = stateStm.executeQuery();
            if(results.next()){
               int state = results.getInt(1);
                if (RegistryConstants.ACTIVE_STATE == state) {
                    return true;
                }
            }
        } catch (SQLException e) {
           log.debug("Inside getResourceStatus trying to access the path: " + resourcePath +
                   " error message" + e.getMessage());
        }
        return false;
    }

    private String getParentPath(String path) {

        String[] parts = path.split(RegistryConstants.PATH_SEPARATOR);
        String parentPath = "";

        if (parts.length > 2) {
            for (int i = 1; i < parts.length - 1; i++) {
                parentPath = parentPath + "/" + parts[i];

            }
        } else {
            parentPath = "/";
        }

        return parentPath;
    }

    /**
     * To get the log messages for all the artifacts
     *
     * @param resourcePath : Path of the resource
     * @param userName     : user name
     * @param from         : from which date
     * @param to           : to date
     * @param conn         : db connection
     * @return List of LogEntry lits
     * @throws RegistryException : if something went wrong
     * @param conn : Connection to the DB
     */
    public List getLogs(String resourcePath,
                        String userName,
                        java.util.Date from,
                        java.util.Date to,
                        Connection conn) throws Exception {
        String SQL = "SELECT UPDATED_TIME , AUTHOR , AID FROM VERSIONS ";
        long aid = 0;
        if (resourcePath != null) {
            ResourceDAO resourceDAO = new ResourceDAO();
            aid = resourceDAO.getArtifactID(resourcePath, conn);
        }
        boolean where = false;
        String whereStatment = "";
        if (aid != 0) {
            whereStatment = "WHERE AID=?";
            where = true;
        }

        if (userName != null) {
            if (where) {
                whereStatment = whereStatment + " AND AUTHOR=?";
            } else {
                whereStatment = "WHERE AUTHOR= ?";
                where = true;
            }
        }
        if (from != null) {
            if (where) {
                whereStatment = whereStatment + " AND UPDATED_TIME > ?";
            } else {
                whereStatment = " WHERE UPDATED_TIME > ?";
                where = true;
            }
        }
        if (to != null) {
            if (where) {
                whereStatment = whereStatment + " AND UPDATED_TIME < ?";
            } else {
                whereStatment = " WHERE UPDATED_TIME < ?";
            }
        }
        SQL = SQL + whereStatment;
        
        PreparedStatement ps = conn.prepareStatement(SQL);
        int count = 1;
        if (aid != 0) {
            ps.setLong(count, aid);
            count++;
        }
        if (userName != null) {
            ps.setString(count, userName);
            count++;
        }
        if (from != null) {
            ps.setTimestamp(count, new Timestamp(from.getTime()));
            count++;
        }
        if (to != null) {
            ps.setTimestamp(count, new Timestamp(to.getTime()));
        }
        ResultSet result = ps.executeQuery();
        ArrayList logevents = new ArrayList();

        while (result.next()) {
            LogEntry logEntry = new LogEntry();
            logEntry.setAction(LogEntry.UPDATE);
            logEntry.setDate(result.getTime("UPDATED_TIME"));
            logEntry.setUserName(result.getString("AUTHOR"));
            aid = result.getLong("AID");
            logEntry.setResourcePath(getResourcePath(aid, conn));
            logevents.add(logEntry);
        }
        return logevents;
    }
}
